# -*- coding: utf-8 -*-

from random import shuffle

# Imports: Zope
from zope.interface import implements
from zope.interface import Attribute
from zope.app.annotation.interfaces import IAttributeAnnotatable

# Imports: Archetypes
from Products.Archetypes.public import BaseContent
from Products.Archetypes.public import registerType

# CMF
from Products.CMFCore.utils import getToolByName

# Imports: ATCT
from Products.ATContentTypes.content.base import ATCTMixin
from Products.ATContentTypes.content.base import updateActions

# Imports: Vaporisation
from Products.Vaporisation.config         import PROJECTNAME
from Products.Vaporisation.interfaces     import IVaporizedCloud
from Products.Vaporisation.interfaces     import ICustomizableCloud

try: 
    from Products.Portletification.interfaces import IPortlet
except ImportError:
    from interfaces import IPortletifiable as IPortlet


class VaporizedCloud( ATCTMixin, BaseContent ):
    """ A cloudy floating bunch of keywords """

    # Tag storage
    keywords = list()
    all_keys = list()
    tagsTree = dict()

    # Weight
    highest  = 0
    lowest   = 0

    # Customization
    steps    = 10
    joint    = True
    limit    = 0
    restrict = tuple()

    # Basics
    allowed_content_types     = None
    allow_discussion          = False
    _at_rename_after_creation = True

    # Z3 implements
    implements( IVaporizedCloud,
                ICustomizableCloud,
                IPortlet,
                IAttributeAnnotatable )

    # Configuration of my type
    portal_type = metal_type    = "VaporizedCloud"
    archetype_name              = "Cloud of keywords"
    content_icon                = "cloud_icon.gif"
    typeDescription             = "A cloud of keywords"
    typeDescMsgId               = "vaporizedcloud_desc"
    global_allow                = True

    schema                      = BaseContent.schema.copy()

    # Actions
    actions = updateActions(ATCTMixin,
        (
        {
        'id'          : 'tags_tree',
        'name'        : 'Update whole cloud',
        'action'      : 'string:${object_url}/steamer/update_tags_tree',
        'permissions' : ('Modify portal content',)
        },
        {
        'id'          : 'prefs',
        'name'        : 'Preferences',
        'action'      : 'string:${object_url}/preferences',
        'permissions' : ('Modify portal content',)
        },
        ))

    # Views
    default_view                = 'cloud_view'
    immediate_view              = 'cloud_view'
    suppl_views                 = tuple()


    def _getStepForTag(self, tag):
        """
        This method will return a tuple that contains :
        - size of the tag to display, in %, for the template
        - The number of occurence of that tag
        The size of the tag is calculated into the inner function
        calculateTagSize, so it's easy to modify.
        If the occurence of the tags is even, they will all be displayed
        at 100%
        """
        def calculateTagSize(step):
            return int(100 + step * 10)
    
        occurence = self.tagsTree[tag]['weight']

        if self.lowest == self.highest:
            return 100, occurence

        lowest  = float(self.lowest)
        highest = float(self.highest)
        
        step = ((float(occurence) - lowest) /
                (highest - lowest) *
                self.steps)
        
        return calculateTagSize(step), occurence

    
    def _getTagsFromTree(self, keywords):
        """
        This method returns a list of dict.
        Each entry consits in a tag, with the keys :
        - name : the name of the keyword, the label
        - weight: the number of occurence of that keyword
        - size : the size in % of the tag to be displayed
        """
        tags = list()

        for keyword in keywords:
            if keyword in self.tagsTree:
                size, weight = self._getStepForTag(keyword)
                tags.append(dict(name=keyword,
                                 weight=weight,
                                 size=size))
        return tags


    def _getConnectionsFor(self, keywords):
        """
        This is heart of the joint navigation.
        It returns a list of keywords.
        This list is computed from the connection keywords have
        between themselves.
        The joint navigation being cumulative, it sorts an eliminates
        keywords via an intersection.
        """
        tags = None
        for keyword in keywords:            
            if keyword in self.tagsTree:
                tag = self.tagsTree[keyword]
                if tags is None:
                    tags = set(tag['connections'])
                else:
                    tags = tags.intersection(set(tag['connections']))
                    
        if tags is None:
            return list()
            
        return self._getTagsFromTree(tuple(tags))


    def _getVaporizedCloudFor(self, subjects=None):
        """
        This returns the keywords to display in the cloud.
        It takes in consideration the given list of keywords.
        """
        if not self.context.joint or not subjects:
            return self._getTagsFromTree(self.keywords)

        return self._getConnectionsFor(subjects)


    def _updateWitnessWeights(self, value):
        """
        In order to calculate the sizes of the tags to be
        displayed, we have to keep track of the fatest and
        skinniest keywords.
        """
        if value > self.highest:
            self.highest = value
        if value < self.lowest:
            self.lowest = value


    def _updateTree(self, keywords):
        """
        This is the main method to build the tags Tree.
        Although it is heavy duty done here, we do it only
        time to time. So, it's merciless.
        """
        tags = self.tagsTree
        for keyword in keywords:
            if keyword in self.tagsTree:
                self.tagsTree[keyword]['weight'] += 1
                addition = [k for k in keywords
                            if k != keyword and
                            k not in self.tagsTree[keyword]['connections']]
                self.tagsTree[keyword]['connections'].extend(addition)
                
            else:
                tag = dict(weight = 1,
                           connections = [k for k in keywords if k != keyword])
                self.tagsTree[keyword] = tag
                
            self._updateWitnessWeights(self.tagsTree[keyword]['weight'])
        self.tagsTree = tags


    def _restrictTree(self):
        """
        This methods takes cares of removing the tags
        beyond the limit fixed by the user.
        To do so, it will build a complementary list
        that is the keywords, sorted by 'weight', the number
        of occurences. In order to sort the keywords, we need
        the already filled tree. It's not very optimal, but
        it is done only once.
        """
        def sort_by_weight(key1, key2):
            return cmp(self.tagsTree[key2]['weight'],
                       self.tagsTree[key1]['weight'])
        
        keywords = self.tagsTree.keys()
        keywords.sort(sort_by_weight)

        if self.limit:
            keywords = keywords[:self.limit]

        for toDelete in self.tagsTree.keys():
            if toDelete not in keywords:
                del self.tagsTree[toDelete]

        self.keywords = keywords
        shuffle(self.keywords)
        

    def _setTree(self):
        """
        Using the catalog, this method creates a blank tree.
        Taking care of keyword restrictions, it builds a full
        structure containing either the keywords and their
        properties, such as number of occurences.
        Keywords are stored as unicodes.
        """        
        catalog  = getToolByName(self, 'portal_catalog')
        putils   = getToolByName(self, 'plone_utils')
        encoding = putils.getSiteEncoding()

        self.tagsTree = dict()

        # First, we get all the keywords used in our portal
        # Then we transform the keywords into unicode objects
        # And we keep an untouched list of keywords (for the form vocabulary)
        subjects = catalog.uniqueValuesFor('Subject')
        self.all_keys = [unicode(k, encoding) for k in subjects]
        self.all_keys.sort()
        self.keywords = [k for k in self.all_keys]
        
        # Now, taking care of the restricted keywords, we build
        # a restricted list and query all the objects using them,
        # via the portal catalog.
        keywords = [k.encode(encoding) for k in self.keywords
                    if k not in self.restrict]
        objects  = catalog(Subject = keywords)

        # Using the main method, we build the references between the tags.
        # We build a list of tags for each objects, verifying the restriction.
        for obj in objects:
            self._updateTree([unicode(k, encoding) for k in obj.Subject
                              if k in keywords])

        if self.limit or len(self.restrict):
            self._restrictTree()
        

registerType( VaporizedCloud,  PROJECTNAME )
